"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ServerContainer = void 0;
const path_1 = __importDefault(require("path"));
const fs_1 = __importDefault(require("fs"));
const promises_1 = __importDefault(require("fs/promises"));
const vm_1 = __importDefault(require("vm"));
const color_1 = __importDefault(require("@oclif/color"));
const dotenv_1 = __importDefault(require("@cubejs-backend/dotenv"));
const semver_1 = require("semver");
const shared_1 = require("@cubejs-backend/shared");
const native_1 = require("@cubejs-backend/native");
const utils_1 = require("./utils");
const server_1 = require("../server");
function safetyParseSemver(version) {
    if (version) {
        return (0, semver_1.parse)(version);
    }
    return null;
}
class ServerContainer {
    constructor(configuration) {
        this.configuration = configuration;
        this.isCubeConfigEmpty = true;
    }
    getTypeScriptCompiler() {
        if ((0, shared_1.packageExists)('typescript', (0, shared_1.isDockerImage)())) {
            // eslint-disable-next-line global-require
            const { TypescriptCompiler } = require('./typescript-compiler');
            return new TypescriptCompiler();
        }
        throw new Error('Typescript dependency not found. Please run this command from project directory.');
    }
    compareBuiltInAndUserVersions(builtInVersion, userVersion) {
        const compareResult = (0, semver_1.compare)(builtInVersion, userVersion);
        if (this.configuration.debug) {
            console.log('[runProjectDockerDiagnostics] compare', {
                builtIn: builtInVersion.raw,
                user: userVersion.raw,
                compare: compareResult,
            });
        }
        if (compareResult === -1) {
            console.log(`${color_1.default.yellow('warning')} You are using old Docker image (${(0, utils_1.getMajorityVersion)(builtInVersion, true)}) `
                + `with new packages (${(0, utils_1.getMajorityVersion)(userVersion, true)})`);
        }
        if (compareResult === 1) {
            console.log(`${color_1.default.yellow('warning')} You are using old Cube.js packages (${(0, utils_1.getMajorityVersion)(userVersion, true)}) `
                + `with new Docker image (${(0, utils_1.getMajorityVersion)(builtInVersion, true)})`);
        }
    }
    async parseLock() {
        const hasNpm = fs_1.default.existsSync(path_1.default.join(process.cwd(), 'package-lock.json'));
        const hasYarn = fs_1.default.existsSync(path_1.default.join(process.cwd(), 'yarn.lock'));
        if (this.configuration.debug) {
            console.log('[parseLock] do', {
                hasNpm,
                hasYarn
            });
        }
        if (hasNpm && hasYarn) {
            console.log(`${color_1.default.yellow('warning')} You are using two different lock files, both for npm/yarn. Please use only one.`);
            return null;
        }
        if (hasNpm) {
            return (0, utils_1.parseNpmLock)();
        }
        if (hasYarn) {
            return (0, utils_1.parseYarnLock)();
        }
        // @todo Error or notice?
        return null;
    }
    async runProjectDockerDiagnostics(manifest, lock) {
        if (this.configuration.debug) {
            console.log('[runProjectDockerDiagnostics] do');
        }
        const builtInCoreVersion = safetyParseSemver(await (0, shared_1.resolveBuiltInPackageVersion)('@cubejs-backend/server'));
        if (!builtInCoreVersion) {
            return;
        }
        const userCoreVersion = safetyParseSemver(lock.resolveVersion('@cubejs-backend/server'));
        if (userCoreVersion) {
            this.compareBuiltInAndUserVersions(builtInCoreVersion, userCoreVersion);
            return;
        }
        if (manifest.devDependencies) {
            /**
             * It's needed to detect case when user didnt install @cubejs-backend/server, but
             * install @cubejs-backend/postgres-driver and it doesn't fit to built-in server
             */
            const depsToCompareVersions = Object.keys(manifest.devDependencies).filter(utils_1.isCubeNotServerPackage);
            // eslint-disable-next-line no-restricted-syntax
            for (const pkgName of depsToCompareVersions) {
                const pkgVersion = safetyParseSemver(lock.resolveVersion(pkgName));
                if (pkgVersion) {
                    this.compareBuiltInAndUserVersions(builtInCoreVersion, pkgVersion);
                    return;
                }
            }
        }
    }
    async runProjectDiagnostics() {
        if (!fs_1.default.existsSync(path_1.default.join(process.cwd(), 'package.json'))) {
            if (this.configuration.debug) {
                console.log('[runProjectDiagnostics] Unable to find package.json, configuration diagnostics skipped');
            }
            return;
        }
        // eslint-disable-next-line global-require,import/no-dynamic-require
        const manifest = require(path_1.default.join(process.cwd(), 'package.json'));
        if (!manifest) {
            return;
        }
        const deepsToIgnore = [
            '@cubejs-backend/databricks-jdbc-driver',
        ];
        if (manifest.dependencies) {
            // eslint-disable-next-line no-restricted-syntax
            for (const [pkgName] of Object.entries(manifest.dependencies)) {
                if ((0, utils_1.isDevPackage)(pkgName) && !deepsToIgnore.includes(pkgName)) {
                    (0, shared_1.displayCLIWarning)(`"${pkgName}" package must be installed in devDependencies`);
                }
            }
        }
        const lock = await this.parseLock();
        if (!lock) {
            return;
        }
        if (manifest.devDependencies) {
            const coreVersion = safetyParseSemver(lock.resolveVersion('@cubejs-backend/server'));
            if (coreVersion) {
                const depsToCompareVersions = Object.keys(manifest.devDependencies).filter(utils_1.isCubeNotServerPackage);
                // eslint-disable-next-line no-restricted-syntax
                for (const pkgName of depsToCompareVersions) {
                    const pkgVersion = safetyParseSemver(lock.resolveVersion(pkgName));
                    if (pkgVersion && !(0, utils_1.isSimilarPackageRelease)(pkgVersion, coreVersion)) {
                        console.log(`${color_1.default.yellow('error')} "${pkgName}" (${(0, utils_1.getMajorityVersion)(pkgVersion)}) `
                            + `is using another release then @cubejs-backend/server (${(0, utils_1.getMajorityVersion)(coreVersion)}).`);
                    }
                }
            }
        }
        if ((0, shared_1.isDockerImage)()) {
            await this.runProjectDockerDiagnostics(manifest, lock);
        }
        else if (this.configuration.debug) {
            console.log('[runProjectDockerDiagnostics] skipped');
        }
    }
    async runServerInstance(configuration, embedded = false, isCubeConfigEmpty) {
        if (embedded) {
            process.env.CUBEJS_SCHEDULED_REFRESH_TIMER = 'false';
            configuration.scheduledRefreshTimer = false;
        }
        const server = this.createServer(configuration, { isCubeConfigEmpty });
        if (!embedded) {
            try {
                const { version, port } = await server.listen();
                console.log(`🚀 Cube API server (${version}) is listening on ${port}`);
            }
            catch (e) {
                console.error('Fatal error during server start: ');
                console.error(e.stack || e);
                process.exit(1);
            }
        }
        return server;
    }
    createServer(config, systemOptions) {
        return new server_1.CubejsServer(config, systemOptions);
    }
    async lookupConfiguration(override = false) {
        dotenv_1.default.config({
            override,
            multiline: 'line-breaks'
        });
        const devMode = (0, shared_1.getEnv)('devMode');
        if (devMode) {
            process.env.NODE_ENV = 'development';
        }
        if (fs_1.default.existsSync(path_1.default.join(process.cwd(), 'cube.py'))) {
            console.log(`${color_1.default.yellow('warning')} You are using Python configuration 🐍, which is an experimental feature and may not support all features`);
            const supported = (0, shared_1.isNativeSupported)();
            if (supported !== true) {
                throw new Error(`Native extension is required to load Python configuration. ${supported.reason}. Read more: ` +
                    'https://github.com/cube-js/cube/blob/master/packages/cubejs-backend-native/README.md#supported-architectures-and-platforms');
            }
            if ((0, native_1.isFallbackBuild)()) {
                throw new Error('Unable to load Python configuration because you are using the fallback build of native extension. Read more: ' +
                    'https://github.com/cube-js/cube/blob/master/packages/cubejs-backend-native/README.md#supported-architectures-and-platforms');
            }
            return this.loadConfigurationFromPythonFile();
        }
        if (fs_1.default.existsSync(path_1.default.join(process.cwd(), 'cube.ts'))) {
            return this.loadConfigurationFromMemory(this.getTypeScriptCompiler().compileConfiguration());
        }
        if (fs_1.default.existsSync(path_1.default.join(process.cwd(), 'cube.js'))) {
            return this.loadConfigurationFromFile();
        }
        (0, shared_1.displayCLIWarning)('There is no cube.js file. Continue with environment variables');
        return {};
    }
    async loadConfigurationFromMemory(content) {
        if (this.configuration.debug) {
            console.log('Loaded configuration from memory', content);
        }
        const exports = {};
        const script = new vm_1.default.Script(content, {
            displayErrors: true,
        });
        script.runInNewContext({
            require,
            console,
            // Workaround for ES5 exports
            exports
        }, {
            filename: 'cube.js',
        });
        if (exports.default) {
            return exports.default;
        }
        throw new Error('Configure file must export configuration as default.');
    }
    async loadConfigurationFromPythonFile() {
        const content = await promises_1.default.readFile(path_1.default.join(process.cwd(), 'cube.py'), 'utf-8');
        if (this.configuration.debug) {
            console.log('Loaded python configuration file', content);
        }
        return this.loadConfigurationFromPythonMemory(content);
    }
    async loadConfigurationFromPythonMemory(content) {
        const config = await (0, native_1.pythonLoadConfig)(content, {
            fileName: 'cube.py',
        });
        return config;
    }
    async loadConfigurationFromFile() {
        var _a;
        const file = await (_a = path_1.default.join(process.cwd(), 'cube.js'), Promise.resolve().then(() => __importStar(require(_a))));
        if (this.configuration.debug) {
            console.log('Loaded js configuration file', file);
        }
        if (file.default) {
            return file.default;
        }
        throw new Error('Configuration file must export the configuration as default.');
    }
    /**
     * @param embedded Cube.js will start without https/ws/graceful shutdown + without timers
     */
    async start(embedded = false) {
        const makeInstance = async (override) => {
            const userConfig = await this.lookupConfiguration(override);
            const configuration = {
                // By default graceful shutdown is disabled, but this value is needed for reboot
                gracefulShutdown: (0, shared_1.getEnv)('gracefulShutdown') || (process.env.NODE_ENV === 'production' ? 30 : 2),
                ...userConfig,
            };
            const server = await this.runServerInstance(configuration, embedded, Object.keys(userConfig).length === 0);
            return {
                configuration,
                gracefulEnabled: !!((0, shared_1.getEnv)('gracefulShutdown') || userConfig.gracefulShutdown),
                server
            };
        };
        let instance = await makeInstance(false);
        if (!embedded) {
            let shutdownHandler = null;
            let killSignalCount = 0;
            const signalToShutdown = [
                // Signal Terminate - graceful shutdown in Unix systems
                'SIGTERM',
                // Ctrl+C
                'SIGINT'
            ];
            // eslint-disable-next-line no-restricted-syntax
            for (const bindSignal of signalToShutdown) {
                // eslint-disable-next-line no-loop-func
                process.on(bindSignal, async (signal) => {
                    killSignalCount++;
                    if (killSignalCount === 3) {
                        console.log('Received killing signal 3 times, exiting immediately');
                        // 130 is the default exit code when killed by a signal.
                        process.exit(130);
                    }
                    if (instance.gracefulEnabled) {
                        if (shutdownHandler) {
                            return;
                        }
                        console.log(`Received ${signal} signal, shutting down in ${instance.configuration.gracefulShutdown}s`);
                        try {
                            shutdownHandler = instance.server.shutdown(signal, true);
                            process.exit(await shutdownHandler);
                        }
                        catch (e) {
                            console.log(e);
                            process.exit(1);
                        }
                    }
                    else {
                        process.exit(0);
                    }
                });
            }
            let restartHandler = null;
            process.addListener('SIGUSR1', async (signal) => {
                console.log(`Received ${signal} signal, reloading in ${instance.configuration.gracefulShutdown}s`);
                if (restartHandler) {
                    console.log('Unable to restart server while it\'s already restarting');
                    return;
                }
                try {
                    restartHandler = instance.server.shutdown(signal, true);
                    await restartHandler;
                }
                finally {
                    restartHandler = null;
                }
                instance = await makeInstance(true);
            });
        }
    }
}
exports.ServerContainer = ServerContainer;
//# sourceMappingURL=container.js.map